/*
 * flotilla.js - 2016-09-24 1512 UTC
 * This file is part of the flotilla project.
 * Copyright © 2015-2016 kexter. All rights reserved.
 */
/*global project*/
/*jshint sub:true*/
(function () {
  "use strict";
  var fontScale = 1.33,
    saves = [],
    content = document.getElementById("Content"),
    overlay = document.getElementById("Overlay"),
    dialogueDivs = [],
    dialogueNameDivs = [],
    waitTimer = null,
    bgmElem = null,
    soundElem = null,
    vocalElem = null,
    videoElem = null,
    actionPointer = 0,
    actionList = [],
    variables = {},
    currentBGM = null,
    currentChapter = 0,
    currentPage = 0,
    canvasSize = [800, 600],
    defaultTextSpeed = 40,
    enableSkipping = true,
    enableAutoplay = false,
    audioMuted = false,
    skipKey = 308,
    stepKey = 32,
    skipSpeed = 8,
    skipMode = 0,
    autoKey = 65,
    autoDelay = 1,
    masterVolume = 1,
    bgmVolume = 1,
    effectVolume = 1,
    vocalVolume = 1,
    videoVolume = 1;
  function serializeSaves() {
    var d,
      uid = "flotilla:" + project["_uid"],
      data = JSON.stringify(saves);
    if (typeof localStorage !== "undefined") {
      localStorage.setItem(uid, data);
    } else {
      d = new Date();
      if (data !== "") {
        d.setFullYear(d.getFullYear() + 10);
      } else {
        d.setFullYear(d.getFullYear() - 1);
      }
      document.cookie = uid + "=" + data +
        "; expires=" + d.toUTCString() + "; path=/";
    }
  }
  function deserializeSaves() {
    var i, str, strings,
      data = "",
      uid = "flotilla:" + project["_uid"];
    if (typeof localStorage !== "undefined") {
      data = localStorage.getItem(uid);
    } else {
      uid += "=";
      strings = document.cookie.split(";");
      for (i = 0; i < strings.length; i++) {
        str = strings[i];
        while (str.charAt(0) === " ") {
          str = str.substring(1, str.length);
        }
        if (str.indexOf(uid) === 0) {
          data = str.substring(uid.length, str.length);
          break;
        }
      }
    }
    data = !data ? "[]" : data;
    saves = JSON.parse(data);
  }
  function saveState(name) {
    var uid = project.Story.Chapter[currentChapter].Page[currentPage]["_uid"];
    saves.push([
      name,
      uid,
      actionPointer,
      JSON.stringify(variables),
      JSON.stringify(currentBGM)
    ]);
    serializeSaves();
  }
  function loadState(index) {
    var i, save, keys, storedVars;
    if (index > -1 && index < saves.length) {
      stopBGM();
      stopSound();
      stopVocal();
      variables = {};
      save = saves[index];
      showPageByUID(save[1]);
      save[2] = parseInt(save[2], 10);
      if (save[2] > 1) {
        actionPointer = save[2] - 1;
        executeNextAction();
      }
      // Clear the wait-timer, just in case.
      clearTimeout(waitTimer);
      // Copy and set all variables from the save.
      storedVars = JSON.parse(save[3]);
      keys = Object.keys(storedVars);
      variables = {};
      for (i = 0; i < keys.length; i++) {
        // Make sure all variables are integers.
        variables[keys[i]] = parseInt(storedVars[keys[i]], 10);
      }
      playBGM(JSON.parse(save[4]));
    }
  }
  function trySaveState() {
    var name, defaultName;
    defaultName = project.Story.Chapter[currentChapter]._name + " - " +
      project.Story.Chapter[currentChapter].Page[currentPage]._name;
    // Prompt the player for a name for the save.
    name = window.prompt("Please enter a name for your save", defaultName);
    // If the player did not cancel the operation then proceed to save the game. 
    if (name !== null) {
      saveState(name);
    }
  }
  function showSaveMenu() {
    var i, el, container,
      wrapper = document.createElement("div"),
      listDiv = document.createElement("div"),
      backBtn = document.createElement("div");
    function deleteOnClickEvent(e) {
      var index = e.target.getAttribute("data-index");
      if (!!index) {
        e.stopPropagation();
        index = parseInt(index, 10);
        saves.splice(index, 1);
        serializeSaves();
        wrapper.parentNode.removeChild(wrapper);
        showSaveMenu();
      }
    }
    function saveOnClickEvent(e) {
      var index = e.currentTarget.getAttribute("data-index");
      if (!!index) {
        e.stopPropagation();
        index = parseInt(index, 10);
        saves.splice(index, 1);
        trySaveState();
        wrapper.parentNode.removeChild(wrapper);
        showSaveMenu();
      }
    }
    wrapper.className = "save-load-menu";
    wrapper.addEventListener("click", function (e) {
      e.stopPropagation();
    });
    // Add the list-container.
    listDiv.className = "save-load-list";
    wrapper.appendChild(listDiv);
    // Add the "create new save" option.
    container = document.createElement("div");
    container.className = "container";
    el = document.createElement("div");
    el.className = "item new";
    el.innerHTML = "<div>[Create new save]</div>";
    el.addEventListener("click", function (e) {
      e.stopPropagation();
      trySaveState();
      wrapper.parentNode.removeChild(wrapper);
      showSaveMenu();
    });
    container.appendChild(el);
    listDiv.appendChild(container);
    // List all existing saves.
    for (i = saves.length - 1; i >= 0; i--) {
      container = document.createElement("div");
      container.className = "container";
      // Add the delete button.
      el = document.createElement("div");
      el.className = "delete";
      el.innerHTML = "×";
      el.setAttribute("data-index", i);
      el.addEventListener("click", deleteOnClickEvent);
      container.appendChild(el);
      // Add the actual save item.
      el = document.createElement("div");
      el.className = "item";
      el.innerHTML = "<div>#" + (i + 1) + " - " + saves[i][0] + "</div>";
      el.setAttribute("data-index", i);
      el.addEventListener("click", saveOnClickEvent);
      container.appendChild(el);
      listDiv.appendChild(container);
    }
    // Add the back-button.
    backBtn.className = "back-button";
    backBtn.innerHTML = "Back";
    backBtn.addEventListener("click", function (e) {
      e.stopPropagation();
      wrapper.parentNode.removeChild(wrapper);
    });
    wrapper.appendChild(backBtn);
    content.appendChild(wrapper);
  }
  function showLoadMenu() {
    var i, el, container,
      wrapper = document.createElement("div"),
      listDiv = document.createElement("div"),
      backBtn = document.createElement("div");
    function deleteOnClickEvent(e) {
      var index = e.target.getAttribute("data-index");
      if (!!index) {
        e.stopPropagation();
        index = parseInt(index, 10);
        saves.splice(index, 1);
        serializeSaves();
        wrapper.parentNode.removeChild(wrapper);
        showLoadMenu();
      }
    }
    function loadOnClickEvent(e) {
      var index = e.currentTarget.getAttribute("data-index");
      if (!!index) {
        e.stopPropagation();
        index = parseInt(index, 10);
        loadState(index);
      }
    }
    wrapper.className = "save-load-menu";
    wrapper.addEventListener("click", function (e) {
      e.stopPropagation();
    });
    // Add the list-container.
    listDiv.className = "save-load-list";
    wrapper.appendChild(listDiv);
    // List all existing saves.
    for (i = saves.length - 1; i >= 0; i--) {
      container = document.createElement("div");
      container.className = "container";
      // Add the delete button.
      el = document.createElement("div");
      el.className = "delete";
      el.innerHTML = "×";
      el.setAttribute("data-index", i);
      el.addEventListener("click", deleteOnClickEvent);
      container.appendChild(el);
      // Add the actual save item.
      el = document.createElement("div");
      el.className = "item";
      el.innerHTML = "<div>#" + (i + 1) + " - " + saves[i][0] + "</div>";
      el.setAttribute("data-index", i);
      el.addEventListener("click", loadOnClickEvent);
      container.appendChild(el);
      listDiv.appendChild(container);
    }
    // Add the back-button.
    backBtn.className = "back-button";
    backBtn.innerHTML = "Back";
    backBtn.addEventListener("click", function (e) {
      e.stopPropagation();
      wrapper.parentNode.removeChild(wrapper);
    });
    wrapper.appendChild(backBtn);
    content.appendChild(wrapper);
  }
  function convertColor(color) {
    var r, g, b, a;
    if (color.length > 7) {
      color = color.toLowerCase().split("");
      a = parseInt(color[1] + color[2], 16) / 255;
      r = parseInt(color[3] + color[4], 16);
      g = parseInt(color[5] + color[6], 16);
      b = parseInt(color[7] + color[8], 16);
      return "rgba(" + [r, g, b, a].join(",") + ")";
    }
    return color;
  }
  function decomposeColor(color) {
    var r, g, b, a;
    if (color.length > 7) {
      color = color.toLowerCase().split("");
      a = parseInt(color[1] + color[2], 16) / 255;
      r = parseInt(color[3] + color[4], 16);
      g = parseInt(color[5] + color[6], 16);
      b = parseInt(color[7] + color[8], 16);
      return ["rgb(" + [r, g, b].join(",") + ")", a];
    }
    return [color, 1];
  }
  function cleanText(text) {
    text = text || "";
    text = text.replace(/\n/g, "<br>");
    text = text.replace(/ {2}/g, "&nbsp;&nbsp;");
    return text;
  }
  function clearExtraObjects() {
    while (content.firstChild) {
      content.removeChild(content.firstChild);
    }
    videoElem = null;
    dialogueDivs = [];
    dialogueNameDivs = [];
  }
  function expandVariablesInText() {
    var i, j, text, vars, variable, value,
      list = document.getElementsByClassName("text-object");
    for (i = 0; i < list.length; i++) {
      text = list[i].innerHTML;
      vars = text.match(/\[\$.+\]/g);
      if (!!vars) {
        for (j = 0; j < vars.length; j++) {
          variable = vars[j].substr(2, vars[j].length - 3);
          value = typeof variables[variable] === "undefined" ?
            0 : variables[variable];
          text = text.replace(vars[j], "<span class=\"dynamic-text\" " +
            "data-variable=\"" + variable + "\">" + value + "</span>");
        }
        list[i].innerHTML = text;
      }
    }
  }
  function updateDynamicText() {
    var i, variable, value,
      list = document.getElementsByClassName("dynamic-text");
    for (i = 0; i < list.length; i++) {
      variable = list[i].getAttribute("data-variable");
      value = typeof variables[variable] === "undefined" ?
        0 : variables[variable];
      list[i].innerHTML = value;
    }
  }
  function getTexture(id) {
    var i,
      result = "";
    for (i = 0; i < project.Resources.Texture.length; i++) {
      if (project.Resources.Texture[i]["_name"] === id) {
        result = project.Resources.Texture[i]["_source"];
        break;
      }
    }
    return result;
  }
  function addImageObject($, parent) {
    var el, img, src, size, pivot, position, color, opacity, clip,
      scale = typeof $["_scale"] === "undefined" ? 1 : parseFloat($["_scale"]),
      id = $["_texture"];
    parent = typeof parent === "undefined" ? content : parent;
    color = typeof $["_color"] === "undefined" ? "#fff" : $["_color"];
    position = typeof $["_position"] === "undefined" ?
      [0, 0] : $["_position"].split(",");
    size = typeof $["_size"] === "undefined" ? [] : $["_size"].split(",");
    pivot = typeof $["_pivot"] === "undefined" ?
      [0, 0] : $["_pivot"].split(",");
    position[0] -= pivot[0];
    position[1] -= pivot[1];
    color = decomposeColor(color);
    opacity = color[1];
    color = color[0];
    src = getTexture(id);
    if (!!src || id === "null") {
      el = document.createElement("div");
      el.className = "image-object";
      el.setAttribute("data-scale", scale);
      if (typeof $["_name"] !== "undefined") {
        el.setAttribute("data-name", $["_name"]);
      }
      if (typeof $["_animation"] !== "undefined") {
        // TODO: Properly emulate animations.
        // Until properly emulated, only add known animations.
        if ($["_animation"] === "anim_animated-rain") {
          el.className += " fake-anim-animated-rain";
        }
      }
      if (typeof $["_shape"] !== "undefined") {
        if ($["_shape"] === "Circle") {
          el.style.borderRadius = "50%";
        }
      }
      el.style.left = position[0] + "px";
      el.style.top = position[1] + "px";
      if (size.length > 1) {
        el.style.width = size[0].indexOf("%") > -1 ?
          size[0] : size[0] * scale + "px";
        el.style.height = size[1].indexOf("%") > -1 ?
          size[1] : size[1] * scale + "px";
      } else if (id !== "null") {
        // Load the image to get the size of it.
        img = document.createElement("img");
        img.addEventListener("load", function (e) {
          size = [
            e.target.naturalWidth * scale,
            e.target.naturalHeight * scale
          ];
          el.style.width = size[0] + "px";
          el.style.height = size[1] + "px";
        });
        img.src = src;
      }
      if (typeof $["_clip"] !== "undefined") {
        clip = $["_clip"].split(",");
        clip[0] *= -1;
        clip[1] *= -1;
        el.style.backgroundPosition = clip[0] + "px " + clip[1] + "px";
        el.style.backgroundSize = "auto";
      }
      if (id === "null") {
        el.style.backgroundColor = color;
      } else {
        el.style.backgroundImage = "url('" + src + "')";
      }
      if (opacity < 1) {
        el.style.opacity = opacity;
      }
      parent.appendChild(el);
    }
  }
  function addTextObject($, parent) {
    var el, div, color, scale, size, position, alignY,
      text = cleanText($["_text"]),
      isDialogueDiv = false,
      isDialogueNameDiv = false;
    parent = typeof parent === "undefined" ? content : parent;
    if (typeof $.Traits !== "undefined") {
      isDialogueDiv = typeof $.Traits.ShowDialogue !== "undefined" ?
        true : isDialogueDiv;
      isDialogueNameDiv = typeof $.Traits.ShowName !== "undefined" ?
        true : isDialogueNameDiv;
    }
    size = typeof $["_size"] === "undefined" ? [] : $["_size"].split(",");
    color = typeof $["_color"] === "undefined" ? "#fff" : $["_color"];
    scale = typeof $["_scale"] === "undefined" ? 1 : parseFloat($["_scale"]);
    position = typeof $["_position"] === "undefined" ?
      [0, 0] : $["_position"].split(",");
    el = document.createElement("div");
    el.className = "text-object";
    if (typeof $["_name"] !== "undefined") {
      el.setAttribute("data-name", $["_name"]);
    }
    el.style.color = color;
    if (size.length > 1) {
      el.style.width = size[0].indexOf("%") > -1 ?
        size[0] : size[0] * scale + "px";
      el.style.height = size[1].indexOf("%") > -1 ?
        size[1] : size[1] * scale + "px";
    }
    el.style.left = position[0] + "px";
    el.style.top = position[1] + "px";
    el.style.fontFamily = $["_font"];
    el.style.fontStyle = $["_italic"] === "true" ? "italic" : "normal";
    el.style.fontWeight = $["_bold"] === "true" ? "bold" : "normal";
    el.style.fontSize = Math.floor(
      parseInt($["_font-size"], 10) * scale * fontScale) + "px";
    if (typeof $["_wrap"] !== "undefined" && $["_wrap"] === "false") {
      el.style.whiteSpace = "nowrap";
    }
    if (typeof $["_align-x"] !== "undefined") {
      el.style.textAlign = $["_align-x"];
    }
    alignY = typeof $["_align-y"] === "undefined" ? "initial" : $["_align-y"];
    alignY = alignY === "center" ? "middle" : alignY;
    div = document.createElement("div");
    div.style.verticalAlign = alignY;
    div.appendChild(el);
    div.innerHTML = text;
    el.appendChild(div);
    parent.appendChild(el);
    if (isDialogueDiv) {
      dialogueDivs.push(div);
    }
    if (isDialogueNameDiv) {
      dialogueNameDivs.push(div);
    }
  }
  function addButtonObject($, parent) {
    var el, action, screen,
      name = $["_name"],
      position = $["_position"].split(","),
      size = $["_size"].split(","),
      visible = typeof $["_visible"] === "undefined" ?
        true : $["_visible"] === "true",
      scale = typeof $["_scale"] === "undefined" ? 1 : parseFloat($["_scale"]),
      enabled = $["_enabled"] === "true";
    parent = typeof parent === "undefined" ? content : parent;
    function mouseEnterEvent(e) {
      e.stopPropagation();
      e.target.removeEventListener("mouseenter", mouseEnterEvent);
      e.target.innerHTML = "";
      addLayer(enabled ? $.Hover : $.Disabled, e.target);
      e.target.removeEventListener("mouseleave", mouseLeaveEvent);
      e.target.addEventListener("mouseleave", mouseLeaveEvent);
    }
    function mouseLeaveEvent(e) {
      e.stopPropagation();
      e.target.removeEventListener("mouseleave", mouseLeaveEvent);
      e.target.innerHTML = "";
      addLayer(enabled ? $.Default : $.Disabled, e.target);
      e.target.removeEventListener("mouseenter", mouseEnterEvent);
      e.target.addEventListener("mouseenter", mouseEnterEvent);
    }
    el = document.createElement("div");
    el.className = "button-object";
    addLayer(enabled ? $.Default : $.Disabled, el);
    el.addEventListener("mouseenter", mouseEnterEvent);
    el.addEventListener("mouseleave", mouseLeaveEvent);
    el.setAttribute("data-name", name);
    el.style.width = (size[0] * scale) + "px";
    el.style.height = (size[1] * scale) + "px";
    el.style.left = position[0] + "px";
    el.style.top = position[1] + "px";
    if (!visible) {
      el.style.display = "none";
    }
    if (typeof $["_action"] !== "undefined") {
      action = $["_action"];
      if (action === "nothing") {
        // Do nothing.
        el.addEventListener("click", function (e) {
          e.stopPropagation();
        });
      } else if (action === "goto") {
        el.setAttribute("data-goto", $["_page-uid"]);
        el.addEventListener("click", function (e) {
          var target = e.currentTarget.getAttribute("data-goto");
          if (!!target) {
            e.stopPropagation();
            showPageByUID(target);
          }
        });
      } else if (action === "exit") {
        el.addEventListener("click", function (e) {
          e.stopPropagation();
          startGame();
        });
      } else if (action === "open-screen") {
        screen = $["_screen"];
        // Divert known save and load screens to the built-in menus.
        if (screen === "Load 1") {
          el.addEventListener("click", function (e) {
            e.stopPropagation();
            showLoadMenu();
          });
        } else if (screen === "Save 1") {
          el.addEventListener("click", function (e) {
            e.stopPropagation();
            showSaveMenu();
          });
        } else {
          // Handle all other screens normally.
          el.setAttribute("data-screen", screen);
          el.addEventListener("click", function (e) {
            var target = e.currentTarget.getAttribute("data-screen");
            if (!!target) {
              e.stopPropagation();
              showScreen(target);
            }
          });
        }
      } else if (action === "close-screen") {
        el.addEventListener("click", function (e) {
          overlay.style.display = "none";
          content.style.display = "block";
          overlay.innerHTML = "";
        });
      } else {
        console.log("Unkown action: " + action);
        el.addEventListener("click", function (e) {
          e.stopPropagation();
        });
      }
    }
    parent.appendChild(el);
  }
  function addHotspotObject($, parent) {
    var el, action,
      position = $["_position"].split(","),
      scale = parseFloat($["_scale"]),
      pivot = $["_pivot"].split(","),
      size = $["_size"].split(","),
      //visible = $["_visible"] === "true",
      //color = convertColor($["_color"]),
      onclick = $.OnClick;
    parent = typeof parent === "undefined" ? content : parent;
    position[0] -= pivot[0];
    position[1] -= pivot[1];
    size[0] *= scale;
    size[1] *= scale;
    el = document.createElement("div");
    el.className = "hotspot-object";
    if (typeof $["_name"] !== "undefined") {
      el.setAttribute("data-name", $["_name"]);
    }
    el.style.width = size[0] + "px";
    el.style.height = size[1] + "px";
    el.style.left = position[0] + "px";
    el.style.top = position[1] + "px";
    //if (visible) {
    //  el.style.backgroundColor = color;
    //}
    action = onclick["_action"];
    if (action === "nothing") {
      // Do nothing.
    } else if (action === "goto") {
      el.setAttribute("data-goto", onclick["_page-uid"]);
      el.addEventListener("click", function (e) {
        e.stopPropagation();
        showPageByUID(e.target.getAttribute("data-goto"));
      });
    } else if (action === "open-screen") {
      el.setAttribute("data-screen", onclick["_screen"]);
      el.addEventListener("click", function (e) {
        e.stopPropagation();
        showScreen(e.target.getAttribute("data-screen"));
      });
    } else if (action === "exit") {
      el.addEventListener("click", function (e) {
        e.stopPropagation();
        startGame();
      });
    } else {
      console.log("Unknown action: " + action);
    }
    parent.appendChild(el);
  }
  function addBorderObject($, parent) {
    var canvas, ctx, img, color, opacity,
      texture = $["_texture"],
      position = typeof $["_position"] === "undefined" ?
        [0, 0] : $["_position"].split(","),
      size = $["_size"].split(",");
    parent = typeof parent === "undefined" ? content : parent;
    color = typeof $["_color"] === "undefined" ? "#fff" : $["_color"];
    color = decomposeColor(color);
    opacity = color[1];
    color = color[0];
    // Emulate the 3x3 box stetching where the borders remain the same.
    canvas = document.createElement("canvas");
    canvas.className = "border-object";
    canvas.width = size[0];
    canvas.height = size[1];
    canvas.style.width = size[0] + "px";
    canvas.style.height = size[1] + "px";
    canvas.style.left = position[0] + "px";
    canvas.style.top = position[1] + "px";
    parent.appendChild(canvas);
    ctx = canvas.getContext("2d");
    img = document.createElement("img");
    img.addEventListener("load", function () {
      var w = Math.floor(img.naturalWidth / 2),
        h = Math.floor(img.naturalHeight / 2),
        cw = canvas.width,
        ch = canvas.height;
      // Draw the four corners.
      ctx.drawImage(img, 0, 0, w, h, 0, 0, w, h);
      ctx.drawImage(img, w, 0, w, h, cw - w, 0, w, h);
      ctx.drawImage(img, 0, h, w, h, 0, ch - h, w, h);
      ctx.drawImage(img, w, h, w, h, cw - w, ch - h, w, h);
      // Draw the sides.
      ctx.drawImage(img, 0, h, w, 1, 0, h, w, ch - h * 2);
      ctx.drawImage(img, w, h, w, 1, cw - w, h, w, ch - h * 2);
      ctx.drawImage(img, w, 0, 1, h, w, 0, cw - w * 2, h);
      ctx.drawImage(img, w, h, 1, h, w, ch - h, cw - w * 2, h);
      // Draw the middle.
      ctx.drawImage(img, w, h, 1, 1, w, h, cw - w * 2, ch - h * 2);
      // Tint the whole image with the desired color.
      ctx.globalCompositeOperation = "source-in";
      ctx.fillStyle = color;
      ctx.fillRect(0, 0, cw, ch);
      canvas.style.opacity = opacity;
    });
    img.src = getTexture(texture);
  }
  function makeMenu($) {
    var i, el, action, container,
      options = $.Option,
      menu = $["_menu"];
    options = typeof options.length === "undefined" ? [options] : options;
    if (options.length < 1) {
      return;
    }
    // TODO: Properly emulate menus.
    // Util properly emulated, only add known menus.
    if (menu === "Sample menu") {
      for (i = 0; i < dialogueDivs.length; i++) {
        dialogueDivs[i].innerHTML = "";
      }
      container = document.createElement("div");
      container.className = "menu-object fake-sample-menu";
      container.setAttribute("data-name", menu);
      for (i = 0; i < options.length; i++) {
        el = document.createElement("div");
        el.className = "menu-item";
        el.innerHTML = cleanText(options[i]["_phrase"]);
        if (typeof options[i]["_action"] !== "undefined") {
          action = options[i]["_action"];
          if (action === "nothing") {
            el.addEventListener("click", function (e) {
              e.stopPropagation();
              container.parentNode.removeChild(container);
              executeNextAction();
            });
          } else if (action === "setvar") {
            el.setAttribute("data-action", JSON.stringify(options[i]));
            el.addEventListener("click", function (e) {
              e.stopPropagation();
              doSetVar(JSON.parse(e.target.getAttribute("data-action")));
              container.parentNode.removeChild(container);
              executeNextAction();
            });
          } else if (action === "goto") {
            el.setAttribute("data-goto", options[i]["_page-uid"]);
            el.addEventListener("click", function (e) {
              e.stopPropagation();
              showPageByUID(e.target.getAttribute("data-goto"));
            });
          } else {
            console.log("Unknown action: " + action);
            el.addEventListener("click", function (e) {
              e.stopPropagation();
              executeNextAction();
            });
          }
        }
        container.appendChild(el);
      }
      content.appendChild(container);
    } else {
      console.log("Unknown menu type: " + menu);
    }
  }
  function doSetVar($) {
    var op, value, variable;
    if (typeof $["_op"] !== "undefined") {
      op = $["_op"];
      variable = $["_variable"];
      value = parseInt($["_value"], 10);
      if (typeof variables[variable] === "undefined") {
        variables[variable] = 0;
      }
      if (op === "=") {
        variables[variable] = value;
      } else if (op === "+") {
        variables[variable] += value;
      } else if (op === "-") {
        variables[variable] -= value;
      } else {
        console.log("setvar unknown op: " + op);
      }
      updateDynamicText();
    }
  }
  function handleIf($) {
    var op, value, variable, route, action;
    if ($.Condition["_type"] === "mem") {
      op = $.Condition["_op"];
      variable = $.Condition["_variable"];
      value = parseInt($.Condition["_value"], 10);
      variable = typeof variables[variable] === "undefined" ?
        0 : variables[variable];
      if (op === "==") {
        route = variable === value ? $.Event : $.Else;
      } else if (op === ">=") {
        route = variable >= value ? $.Event : $.Else;
      } else if (op === "<=") {
        route = variable <= value ? $.Event : $.Else;
      } else if (op === ">") {
        route = variable > value ? $.Event : $.Else;
      } else if (op === "<") {
        route = variable < value ? $.Event : $.Else;
      } else {
        console.log("Unknow condition-op: " + op);
      }
      action = route["_action"];
      if (action === "nothing") {
        executeNextAction();
      } else if (action === "goto") {
        showPageByUID(route["_page-uid"]);
      }
    } else {
      console.log("Unknown if-condition-type: " + $.Condition["_type"]);
    }
  }
  function playSoundOrVocal($, soundOrVocal, volumeLevel) {
    var i, el, ext, elem,
      src = "",
      name = $["_name"],
      volume = parseFloat($["_volume"]),
      loop = $["_loop"] === "true";
    for (i = 0; i < project.Resources.Audio.length; i++) {
      if (project.Resources.Audio[i]["_name"] === name) {
        src = project.Resources.Audio[i]["_source"];
        break;
      }
    }
    if (!!src) {
      if (soundOrVocal === "vocal") {
        elem = vocalElem;
        vocalElem = null;
      } else {
        elem = soundElem;
        soundElem = null;
      }
      if (!!elem) {
        elem.pause();
        elem.parentNode.removeChild(elem);
      }
      elem = document.createElement("audio");
      elem.setAttribute("data-volume", volume);
      elem.preload = "auto";
      elem.autoplay = true;
      elem.loop = loop;
      elem.volume = volumeLevel * volume;
      elem.muted = audioMuted;
      ext = src.split(".").pop().toLowerCase();
      ext = ext === "mp3" ? "audio/mpeg" :
        (ext === "ogg" ? "audio/ogg" : "audio/wav");
      el = document.createElement("source");
      el.src = src;
      el.type = ext;
      elem.appendChild(el);
      document.body.appendChild(elem);
      if (soundOrVocal === "vocal") {
        vocalElem = elem;
      } else {
        soundElem = elem;
      }
    }
  }
  function playSound($) {
    playSoundOrVocal($, "sound", effectVolume);
  }
  function stopSound($) {
    if (!!soundElem) {
      soundElem.parentNode.removeChild(soundElem);
      soundElem = null;
    }
  }
  function playVocal($) {
    playSoundOrVocal($, "vocal", vocalVolume);
  }
  function stopVocal($) {
    if (!!vocalElem) {
      vocalElem.parentNode.removeChild(vocalElem);
      vocalElem = null;
    }
  }
  function playBGM($) {
    var i, el, ext, name, volume, loop, src;
    $ = typeof $ === "undefined" ? null : $;
    if ($ !== null) {
      name = $["_name"];
      volume = parseFloat($["_volume"]);
      loop = $["_loop"] === "true";
      stopBGM();
      currentBGM = $;
      src = "";
      for (i = 0; i < project.Resources.Audio.length; i++) {
        if (project.Resources.Audio[i]["_name"] === name) {
          src = project.Resources.Audio[i]["_source"];
          break;
        }
      }
      if (!!src) {
        bgmElem = document.createElement("audio");
        bgmElem.setAttribute("data-volume", volume);
        bgmElem.preload = "auto";
        bgmElem.autoplay = true;
        bgmElem.loop = loop;
        bgmElem.volume = bgmVolume * volume;
        bgmElem.muted = audioMuted;
        ext = src.split(".").pop().toLowerCase();
        ext = ext === "mp3" ? "audio/mpeg" :
          (ext === "ogg" ? "audio/ogg" : "audio/wav");
        el = document.createElement("source");
        el.src = src;
        el.type = ext;
        bgmElem.appendChild(el);
        document.body.appendChild(bgmElem);
      }
    }
  }
  function stopBGM($) {
    $ = typeof $ === "undefined" ? {} : $;
    if (!!bgmElem) {
      bgmElem.pause();
      bgmElem.parentNode.removeChild(bgmElem);
      bgmElem = null;
      currentBGM = null;
    }
  }
  function playVideo($) {
    var el,
      src = $["_filename"],
      volume = parseFloat($["_volume"]),
      loop = $["_loop"] === "true";
    if (!!videoElem) {
      videoElem.parentNode.removeChild(videoElem);
    }
    src = src.replace(/\.wmv$/, ".mp4");
    videoElem = document.createElement("video");
    videoElem.setAttribute("data-name", "Video object");
    videoElem.autoplay = "true";
    videoElem.loop = loop;
    videoElem.volume = volume;
    videoElem.muted = audioMuted;
    el = document.createElement("source");
    el.src = src;
    el.type = "video/mp4";
    videoElem.appendChild(el);
    content.appendChild(videoElem);
  }
  function stopVideo($) {
    $ = typeof $ === "undefined" ? {} : $;
    if (!!videoElem) {
      videoElem.pause();
    }
  }
  function doFade($) {
    var i, el,
      type = $["_fade"],
      time = $["_time"],
      wait = $["_wait"] === "true",
      object = $["_object"];
    el = getObjectByName(object);
    if (!!el) {
      for (i = 0; i < dialogueDivs.length; i++) {
        dialogueDivs[i].innerHTML = "";
      }
      el.style.transitionDuration = time + "s";
      if (type === "out") {
        el.style.opacity = 1;
        el.classList.add("with-transition");
        setTimeout(function () {
          el.style.opacity = 0;
        }, 10);
      } else if (type === "in") {
        el.style.opacity = 0;
        el.classList.add("with-transition");
        setTimeout(function () {
          el.style.opacity = 1;
        }, 10);
      }
    } else {
      console.log("Object not found: " + object);
    }
    if (wait && time > 0) {
      waitTimer = setTimeout(function () {
        if (!!el) {
          el.classList.remove("with-transition");
        }
        executeNextAction();
      }, Math.floor(parseFloat(time) * 1000));
    } else {
      if (!!el) {
        el.classList.remove("with-transition");
      }
      executeNextAction();
    }
  }
  function doMove($) {
    var el,
      time = parseFloat($["_time"]),
      wait = $["_wait"] === "true",
      to = $["_to"].split(","),
      object = $["_object"];
    el = getObjectByName(object);
    if (!!el) {
      if (time > 0) {
        el.style.webkitTransition = "all " + time + "s linear 0s";
        el.style.transition = "all " + time + "s linear 0s";
      }
      el.style.left = to[0] + "px";
      el.style.top = to[1] + "px";
    } else {
      console.log("Object not found: " + object);
    }
    if (wait && time > 0) {
      waitTimer = setTimeout(executeNextAction, Math.floor(time * 1000));
    } else {
      executeNextAction();
    }
  }
  function doScale($) {
    var el, baseScale, size,
      time = parseFloat($["_time"]),
      wait = $["_wait"] === "true",
      scale = $["_to"],
      object = typeof $["_object"] === "undefined" ? null : $["_object"];
    if (!!object) {
      el = getObjectByName(object);
      if (!!el) {
        if (time > 0) {
          el.style.webkitTransition = "all " + time + "s linear 0s";
          el.style.transition = "all " + time + "s linear 0s";
        }
        baseScale = el.getAttribute("data-scale");
        baseScale = baseScale === 0 ? 1 : baseScale;
        el.setAttribute("data-scale", scale);
        size = [el.clientWidth / baseScale * scale,
          el.clientHeight / baseScale * scale];
        el.style.width = size[0] + "px";
        el.style.height = size[1] + "px";
      } else {
        console.log("Object not found: " + object);
      }
    }
    if (wait && time > 0) {
      waitTimer = setTimeout(executeNextAction, Math.floor(time * 1000));
    } else {
      executeNextAction();
    }
  }
  function doTransform($) {
    var el,
      time = parseFloat($["_time"]),
      wait = $["_wait"] === "true",
      object = $["_object"];
    el = getObjectByName(object);
    if (!!el) {
      if (time > 0) {
        el.style.webkitTransition = "all " + time + "s linear 0s";
        el.style.transition = "all " + time + "s linear 0s";
      }
      // TODO: Implement doTransform().
      console.log("doTransform() -> ", $);
    } else {
      console.log("Object not found: " + object);
    }
    if (wait && time > 0) {
      waitTimer = setTimeout(executeNextAction, Math.floor(time * 1000));
    } else {
      executeNextAction();
    }
  }
  function doShowHide($) {
    var el,
      show = $["_show"] === "true",
      wait = $["_wait"] === "true",
      fade = $["_fade"] === "true",
      object = $["_object"],
      time = $["_time"];
    if (fade) {
      doFade({
        _fade: show ? "in" : "out",
        _object: object,
        _time: time
      });
    } else {
      el = getObjectByName(object);
      if (!!el) {
        el.style.opacity = show ? 1 : 0;
      } else {
        console.log("Object not found: " + object);
      }
      if (wait && time > 0) {
        waitTimer = setTimeout(executeNextAction,
          Math.floor(parseFloat(time) * 1000));
      } else {
        executeNextAction();
      }
    }
  }
  function doShowHideLayer($) {
    var el = getObjectByName($["_layer"]),
      show = $["_show"] === "true",
      wait = $["_wait"] === "true",
      time = $["_time"];
    function showHideInternal() {
      if (!!el) {
        el.style.display = show ? "block" : "none";
      } else {
        console.log("Layer not found: " + $["_layer"]);
      }
      executeNextAction();
    }
    if (wait && time > 0) {
      waitTimer = setTimeout(showHideInternal,
        Math.floor(parseFloat(time) * 1000));
    } else {
      showHideInternal();
    }
  }
  function doShake($) {
    content.className = "content fake-shake";
    waitTimer = setTimeout(function () {
      content.className = "content";
      executeNextAction();
    }, Math.floor(parseFloat($["_time"]) * 1000));
  }
  function doMirror($) {
    var el = getObjectByName($["_object"]);
    if (!!el) {
      el.style.transform = "scaleX(-1)";
    }
  }
  function doSay($) {
    var i,
      startNewLine = typeof $["_newline"] === "undefined" ?
        true : $["_newline"] === "true",
      doAppend = typeof $["_append"] === "undefined" ?
        false : $["_append"] === "true",
      who = typeof $["_who"] === "undefined" ? "" : $["_who"],
      text = dialogueDivs.length > 0 ? dialogueDivs[0].innerHTML : "";
    if (!!text && doAppend) {
      text += cleanText((startNewLine ? "\n" : "") + $["_phrase"]);
    } else {
      text = cleanText($["_phrase"]);
    }
    for (i = 0; i < dialogueDivs.length; i++) {
      dialogueDivs[i].innerHTML = text;
    }
    for (i = 0; i < dialogueNameDivs.length; i++) {
      dialogueNameDivs[i].innerHTML = who;
    }
    expandVariablesInText();
    waitForOnclick();
  }
  function waitForOnclickEvent(e) {
    e.stopPropagation();
    content.removeEventListener("click", waitForOnclickEvent);
    executeNextAction();
  }
  function waitForOnclick() {
    content.addEventListener("click", waitForOnclickEvent);
  }
  function handleAction($) {
    var type;
    type = $["_type"];
    if (type === "setvar") {
      doSetVar($);
    } else {
      console.log("Unknown action: " + type);
      console.log($);
    }
  }
  function executeNextAction() {
    var $, type;
    if (actionPointer > actionList.length - 1) {
      actionList = [];
      actionPointer = 0;
      nextPage();
      return;
    }
    $ = actionList[actionPointer];
    actionPointer += 1;
    type = $["_type"];
    if (type === "wait" && ($["_timed"] === "true" ||
        $["_mode"] === "time" || $["_mode"] === "video")) {
      waitTimer = setTimeout(executeNextAction,
        Math.floor(parseFloat($["_time"]) * 1000));
      return;
    } else if (type === "wait" && $["_mode"] === "user") {
      waitForOnclick();
      return;
    } else if (type === "wait" && $["_mode"] === "forever") {
      return;
    } else if (type === "say") {
      if (dialogueDivs.length > 0) {
        doSay($);
        return;
      }
      console.log("No designated dialogue-target is specified.");
    } else if (type === "goto") {
      showPageByUID($["_page-uid"]);
      return;
    } else if (type === "menu") {
      makeMenu($);
      return;
    } else if (type === "exit") {
      startGame();
      return;
    } else if (type === "fade") {
      doFade($);
      return;
    } else if (type === "transform") {
      doTransform($);
      return;
    } else if (type === "move") {
      doMove($);
      return;
    } else if (type === "scale") {
      doScale($);
      return;
    } else if (type === "showhide") {
      doShowHide($);
      return;
    } else if (type === "showhidelayer") {
      doShowHideLayer($);
      return;
    } else if (type === "shake" || type === "shake-scene") {
      doShake($);
      return;
    } else if (type === "setvar") {
      doSetVar($);
    } else if (type === "if") {
      handleIf($);
      return;
    } else if (type === "mirror") {
      doMirror($);
    } else if (type === "playbgm") {
      playBGM($);
    } else if (type === "stopbgm") {
      stopBGM($);
    } else if (type === "playsnd") {
      playSound($);
    } else if (type === "stopsnd") {
      stopSound($);
    } else if (type === "playvocal") {
      playVocal($);
    } else if (type === "stopvocal") {
      stopVocal($);
    } else if (type === "playvideo") {
      playVideo($);
    } else if (type === "stopvideo") {
      stopVideo($);
    } else {
      console.log("Unknown action: " + type);
      console.log($);
    }
    executeNextAction();
  }
  function addLayer($, parent) {
    var i, container, position, size, scale;
    parent = typeof parent === "undefined" ? content : parent;
    if (typeof $["_name"] === "undefined") {
      container = parent;
    } else {
      container = document.createElement("div");
      container.className = "layer";
      container.setAttribute("data-name", $["_name"]);
      parent.appendChild(container);
    }
    position = typeof $["_position"] === "undefined" ?
      [] : $["_position"].split(",");
    size = typeof $["_size"] === "undefined" ? [] : $["_size"].split(",");
    scale = typeof $["_scale"] === "undefined" ? 1 : parseFloat($["_scale"]);
    if (position.length > 1) {
      container.style.left = position[0] + "px";
      container.style.top = position[1] + "px";
    }
    if (size.length > 1) {
      container.style.width = size[0].indexOf("%") > -1 ?
        size[0] : size[0] * scale + "px";
      container.style.height = size[1].indexOf("%") > -1 ?
        size[1] : size[1] * scale + "px";
    }
    // Add image-objects.
    if (typeof $.Image !== "undefined") {
      if (typeof $.Image.length === "undefined") {
        $.Image = [$.Image];
      }
      for (i = 0; i < $.Image.length; i++) {
        addImageObject($.Image[i], container);
      }
    }
    // Add "border"-objects.
    if (typeof $.Border !== "undefined") {
      if (typeof $.Border.length === "undefined") {
        $.Border = [$.Border];
      }
      for (i = 0; i < $.Border.length; i++) {
        addBorderObject($.Border[i], container);
      }
    }
    // Add groups.
    if (typeof $.Group !== "undefined") {
      if (typeof $.Group.length === "undefined") {
        $.Group = [$.Group];
      }
      for (i = 0; i < $.Group.length; i++) {
        addLayer($.Group[i], container);
      }
    }
    // Add text-objects.
    if (typeof $.Text !== "undefined") {
      if (typeof $.Text.length === "undefined") {
        $.Text = [$.Text];
      }
      for (i = 0; i < $.Text.length; i++) {
        addTextObject($.Text[i], container);
      }
    }
    // Add system-text-objects.
    if (typeof $.SystemText !== "undefined") {
      if (typeof $.SystemText.length === "undefined") {
        $.SystemText = [$.SystemText];
      }
      for (i = 0; i < $.SystemText.length; i++) {
        addTextObject($.SystemText[i], container);
      }
    }
    // Add hotspot-objects.
    if (typeof $.Hotspot !== "undefined") {
      if (typeof $.Hotspot.length === "undefined") {
        $.Hotspot = [$.Hotspot];
      }
      for (i = 0; i < $.Hotspot.length; i++) {
        addHotspotObject($.Hotspot[i], container);
      }
    }
    // Add button-objects.
    if (typeof $.Button !== "undefined") {
      if (typeof $.Button.length === "undefined") {
        $.Button = [$.Button];
      }
      for (i = 0; i < $.Button.length; i++) {
        addButtonObject($.Button[i], container);
      }
    }
    // Expand variables in text-objects.
    expandVariablesInText();
  }
  function addLayerFromGlobal(name, parent) {
    var i, layers;
    if (typeof project.Story.Global.Layer.length === "undefined") {
      project.Story.Global.Layer = [project.Story.Global.Layer];
    }
    layers = project.Story.Global.Layer;
    for (i = 0; i < layers.length; i++) {
      if (layers[i]["_name"] === name) {
        addLayer(layers[i], parent);
      }
    }
  }
  function getObjectByName(name) {
    var i,
      result = null,
      list = document.getElementsByTagName("*");
    for (i = 0; i < list.length; i++) {
      if (list[i].getAttribute("data-name") === name) {
        result = list[i];
        break;
      }
    }
    return result;
  }
  function addScene($, parent) {
    var i;
    parent = typeof parent === "undefined" ? content : parent;
    if (typeof $.Layer !== "undefined") {
      if (typeof $.Layer.length === "undefined") {
        $.Layer = [$.Layer];
      }
      for (i = 0; i < $.Layer.length; i++) {
        if (typeof $.Layer[i]["_import"] !== "undefined" &&
            $.Layer[i]["_import"] === "true") {
          addLayerFromGlobal($.Layer[i]["_name"], parent);
        } else {
          addLayer($.Layer[i], parent);
        }
      }
    }
  }
  function showScreen(name) {
    var i, j;
    for (i = 0; i < project.Story.Screens.Screen.length; i++) {
      if (project.Story.Screens.Screen[i]["_name"] === name) {
        addScene(project.Story.Screens.Screen[i].Scene);
        if (typeof project.Story.Screens.Screen[i].Action !== "undefined") {
          if (typeof project.Story.Screens.Screen[i].Action.length ===
              "undefined") {
            project.Story.Screens.Screen[i].Action =
              [project.Story.Screens.Screen[i].Action];
          }
          for (j = 0; j < project.Story.Screens.Screen[i].Action.length; j++) {
            handleAction(project.Story.Screens.Screen[i].Action[j]);
          }
        }
        break;
      }
    }
  }
  function showPageByUID(uid) {
    var i, j;
    for (i = 0; i < project.Story.Chapter.length; i++) {
      for (j = 0; j < project.Story.Chapter[i].Page.length; j++) {
        if (project.Story.Chapter[i].Page[j]["_uid"] === uid) {
          showPage(project.Story.Chapter[i].Page[j], i, j);
          return;
        }
      }
    }
    console.log("No page found by uid: " + uid + ". Moving to the next page.");
    nextPage();
  }
  function showPage($, chapter, page) {
    var src;
    //console.log("Showing: " + $["_name"] + " (uid: " + $["_uid"] + ")");
    currentChapter = chapter;
    currentPage = page;
    clearExtraObjects();
    content.style.display = "block";
    overlay.style.display = "none";
    if (waitTimer !== null) {
      clearTimeout(waitTimer);
      waitTimer = null;
    }
    if (typeof $["_color"] !== "undefined") {
      content.style.backgroundColor = $["_color"];
    } else {
      content.style.backgroundColor = "#222";
    }
    if (typeof $.Scene === "object") {
      addScene($.Scene);
    }
    // Add the bakground image, if specified.
    if (typeof $["_background"] !== "undefined") {
      src = getTexture($["_background"]);
      if (!!src) {
        content.style.backgroundImage = "url('" + src + "')";
      }
    } else {
      content.style.backgroundImage = "";
    }
    // Handle automatic timed page transitions.
    if (typeof $.Action !== "undefined") {
      if (typeof $.Action.length === "undefined") {
        $.Action = [$.Action];
      }
      actionPointer = 0;
      actionList = $.Action;
      executeNextAction();
    } else {
      waitForOnclick();
    }
    // Handle pages with Angelscript-code here, detect based on project-uid.
    if (project["_uid"] === "1401043508") {
      // Wallis and the Time Machine specific code.
      if ($["_uid"] === "10") {
        // The safe-puzzle.
        wallisSafePuzzle();
      }
    } else if (project["_uid"] === "1769811879") {
      // Vera Blanc and the Cursed Mummy.
      if ($["_uid"] === "5901") {
        // The keypad-puzzle.
        veraMummyKeypadPuzzle();
      }
    }
  }
  function nextPage() {
    var chapter = currentChapter,
      page = currentPage + 1;
    if (page > project.Story.Chapter[chapter].Page.length - 1) {
      chapter += 1;
      page = 0;
    }
    showPage(project.Story.Chapter[chapter].Page[page], chapter, page);
  }
  function applySettings($) {
    // Set the settings.
    variables = {};
    canvasSize = $["_canvas-size"].split(",");
    fontScale = typeof $["_fontscale"] === "undefined" ?
      fontScale : parseFloat($["_fontscale"]);
    defaultTextSpeed = parseInt($["_default-text-speed"], 10);
    enableSkipping = $["_enable-skipping"] === "true";
    enableAutoplay = $["_enable-autoplay"] === "true";
    skipKey = parseInt($["_skip-key"], 10);
    stepKey = parseInt($["_step-key"], 10);
    skipSpeed = parseInt($["_skip-speed"], 10);
    skipMode = parseInt($["_skip-mode"], 10);
    autoKey = parseInt($["_auto-key"], 10);
    autoDelay = parseInt($["_auto-delay"], 10);
    // TODO: Readd support for project specific default volume settings.
    //masterVolume = parseFloat($["_master-volume"]);
    //bgmVolume = parseFloat($["_bgm-volume"]);
    //effectVolume = parseFloat($["_effect-volume"]);
    //vocalVolume = parseFloat($["_vocal-volume"]);
    //videoVolume = parseFloat($["_video-volume"]);
    // Initialize the content-area.
    content.className = "content";
    content.style.width = canvasSize[0] + "px";
    content.style.height = canvasSize[1] + "px";
    content.style.backgroundColor = "#000";
    content.style.backgroundImage = "";
    // Initialize the overlay-area.
    overlay.style.width = canvasSize[0] + "px";
    overlay.style.height = canvasSize[1] + "px";
    overlay.style.backgroundColor = "#000";
    overlay.addEventListener("click", function (e) {
      e.stopPropagation();
    });
  }
  function fixProject() {
    var i;
    for (i = 0; i < project.Story.Chapter.length; i++) {
      if (typeof project.Story.Chapter[i].Page.length === "undefined") {
        project.Story.Chapter[i].Page = [project.Story.Chapter[i].Page];
      }
    }
  }
  function startGame() {
    fixProject();
    document.head.getElementsByTagName("title")[0].innerHTML = project["_name"];
    content.removeEventListener("click", waitForOnclickEvent);
    if (waitTimer !== null) {
      clearTimeout(waitTimer);
      waitTimer = null;
    }
    stopBGM();
    stopSound();
    stopVocal();
    applySettings(project.Settings);
    deserializeSaves();
    showPage(project.Story.Chapter[0].Page[0], 0, 0);
  }
  function addScript(src, callback) {
    var el = document.createElement("script");
    el.async = false;
    el.charset = "utf-8";
    if (typeof callback === "function") {
      el.addEventListener("load", callback);
    }
    el.src = src;
    document.body.appendChild(el);
  }
  function addVolumeControlPanel() {
    var masterSlider, sfxSlider, musicSlider,
      volumeControl = document.createElement("div"),
      muteBtn = document.createElement("div");
    function toggleAudio() {
      var i,
        audioTags = document.getElementsByTagName("audio"),
        videoTags = document.getElementsByTagName("video");
      audioMuted = !audioMuted;
      muteBtn.innerHTML = audioMuted ? "🔇" : "🔊";
      for (i = 0; i < audioTags.length; i++) {
        audioTags[i].muted = audioMuted;
      }
      for (i = 0; i < videoTags.length; i++) {
        videoTags[i].muted = audioMuted;
      }
    }
    function createVolumeSlider(label, func) {
      var sliderControl = document.createElement("div"),
        sliderLabel = document.createElement("div"),
        sliderWrapper = document.createElement("div"),
        slider = document.createElement("div");
      function volumeControlOnClick(e) {
        var el = e.currentTarget.getElementsByClassName("fl-volume-slider")[0],
          x = Math.min(e.clientX - e.currentTarget.offsetLeft, 100),
          y = Math.floor(x / 5);
        el.style.borderLeftWidth = x + "px";
        el.style.borderBottomWidth = y + "px";
        el.style.top = (20 - y) + "px";
        if (typeof func === "function") {
          func(x/100);
        }
      }
      sliderLabel.className = "fl-volume-slider-label";
      sliderLabel.innerHTML = label;
      slider.className = "fl-volume-slider";
      sliderWrapper.className = "fl-volume-slider-wrapper";
      sliderWrapper.appendChild(slider);
      sliderWrapper.addEventListener("click", volumeControlOnClick);
      sliderControl.className = "fl-volume-slider-control";
      sliderControl.appendChild(sliderLabel);
      sliderControl.appendChild(sliderWrapper);
      return sliderControl;
    }
    function changeVolume(audioElem, volume) {
      var baseVolume;
      if (audioElem) {
        baseVolume = parseFloat(audioElem.getAttribute("data-volume")) || 1;
        audioElem.volume = volume * baseVolume;
      }
    }
    function changeMasterVolume(volume) {
      // TODO: Implement master volume control.
      masterVolume = volume;
    }
    function changeSfxVolume(volume) {
      effectVolume = volume;
      vocalVolume = volume;
      changeVolume(soundElem, effectVolume);
      changeVolume(vocalElem, vocalVolume);
    }
    function changeMusicVolume(volume) {
      bgmVolume = volume;
      changeVolume(bgmElem, bgmVolume);
    }
    // Add the mute button.
    muteBtn.className = "fl-mute-button";
    muteBtn.innerHTML = "🔊";
    muteBtn.title = "Mute";
    muteBtn.addEventListener("click", toggleAudio);
    // Create the volume sliders.
    masterSlider = createVolumeSlider("Master", changeMasterVolume);
    sfxSlider = createVolumeSlider("SFX", changeSfxVolume);
    musicSlider = createVolumeSlider("Music", changeMusicVolume);
    // Build the volume control panel.
    volumeControl.className = "fl-volume-control";
    volumeControl.appendChild(muteBtn);
    //volumeControl.appendChild(masterSlider);
    volumeControl.appendChild(sfxSlider);
    volumeControl.appendChild(musicSlider);
    document.body.appendChild(volumeControl);
  }
  function loadGame() {
    var el = document.createElement("link");
    // Add the main style-sheet.
    el.rel = "stylesheet";
    el.href = "style.css";
    document.head.appendChild(el);
    // Add the volume control panel.
    addVolumeControlPanel();
    // Load the game data.
    addScript("project.js", startGame);
  }
  // Extra, game-specific scripts.
  function wallisSafePuzzle() {
    var el, safeTimer,
      current = 0,
      rotation = 0;
    function stepInnerLock() {
      var variable = String.fromCharCode("b".charCodeAt(0) + current);
      rotation += 6;
      rotation = rotation >= 360 ? rotation - 360 : rotation;
      el.style.transform = "rotateZ(" + rotation + "deg)";
      doSetVar({
        _variable: variable,
        _op: "=",
        _value: Math.floor(rotation / 6)
      });
      safeTimer = setTimeout(stepInnerLock, 500);
    }
    function checkWinOrReset() {
      if (variables["b"] === 6 &&
          variables["c"] === 1 &&
          variables["d"] === 2 &&
          variables["e"] === 1 &&
          variables["f"] === 5) {
        // Jump to the "WIN!" page.
        showPageByUID("11");
      } else {
        current = 0;
        doSetVar({ _variable: "b", _op: "=", _value: 0 });
        doSetVar({ _variable: "c", _op: "=", _value: 0 });
        doSetVar({ _variable: "d", _op: "=", _value: 0 });
        doSetVar({ _variable: "e", _op: "=", _value: 0 });
        doSetVar({ _variable: "f", _op: "=", _value: 0 });
      }
    }
    el = getObjectByName("Safe lock_inner");
    if (!!el) {
      el.addEventListener("mousedown", function () {
        safeTimer = setTimeout(stepInnerLock, 500);
      });
      el.addEventListener("mouseup", function () {
        clearTimeout(safeTimer);
        rotation = 0;
        el.style.transform = "rotateZ(0deg)";
        current += 1;
        if (current > 4) {
          checkWinOrReset();
        }
      });
    }
  }
  function veraMummyKeypadPuzzle() {
    function setKeypadKeyState(name, enabled) {
      var el = getObjectByName(name);
      if (!!el) {
        el.setAttribute("data-state", enabled ? "1" : "0");
        el.style.backgroundColor = enabled ? "blue" : "coral"; // "#ff7f50"
      }
    }
    function setKeypadKeyClick(name, callback) {
      var el = getObjectByName(name);
      if (!!el) {
        el.addEventListener("click", function () {
          if (typeof callback === "function") {
            callback();
          }
          checkWin();
        });
      }
    }
    function checkWin() {
      var i, el,
        done = true;
      for (i = 1; i < 9; i++) {
        el = getObjectByName("Bar" + i);
        if (!!el) {
          done = done && el.getAttribute("data-state") === "1";
        } else {
          done = false;
          break;
        }
      }
      if (done) {
        showScreen("WIN_SCREEN");
      }
    }
    // Initialize the key to the starting state.
    setKeypadKeyState("Bar1", false);
    setKeypadKeyState("Bar2", false);
    setKeypadKeyState("Bar3", false);
    setKeypadKeyState("Bar4", false);
    setKeypadKeyState("Bar5", false);
    setKeypadKeyState("Bar6", false);
    setKeypadKeyState("Bar7", false);
    setKeypadKeyState("Bar8", false);
    // Add the onclick events.
    setKeypadKeyClick("Bar1", function () {
      setKeypadKeyState("Bar2", true);
      setKeypadKeyState("Bar3", true);
    });
    setKeypadKeyClick("Bar2", function () {
      setKeypadKeyState("Bar4", true);
      setKeypadKeyState("Bar5", false);
    });
    setKeypadKeyClick("Bar3", function () {
      setKeypadKeyState("Bar3", true);
      setKeypadKeyState("Bar5", true);
      setKeypadKeyState("Bar8", true);
    });
    setKeypadKeyClick("Bar4", function () {
      setKeypadKeyState("Bar7", true);
      setKeypadKeyState("Bar6", false);
    });
    setKeypadKeyClick("Bar5", function () {
      setKeypadKeyState("Bar1", true);
    });
    setKeypadKeyClick("Bar6", function () {
      setKeypadKeyState("Bar6", true);
      setKeypadKeyState("Bar3", false);
    });
    setKeypadKeyClick("Bar7", function () {
      setKeypadKeyState("Bar2", true);
      setKeypadKeyState("Bar4", true);
      setKeypadKeyState("Bar6", true);
    });
    setKeypadKeyClick("Bar8", function () {
      setKeypadKeyState("Bar7", true);
    });
  }
  window.addEventListener("load", loadGame);
}());